/* This header file was based on nhmlg.h in NHML (Galtier & Gouy, 1998; Galtier, 2001) */

#define _GNU_SOURCE
#define _XOPEN_SOURCE 600
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <limits.h>
#include <float.h>

#include "portab.h"


#if defined(USE_BLAS) || defined(USE_LAPACK)
// choose a vendor BLAS to use, and let the rest of procov know if we have
// CBLAS and/or CLAPACK, or just the f77 interface (column major only).
// Procov uses the CBLAS interfaces because there are a few different flavours
// of non-cblas interfaces for C.  e.g. char * vs. char for the transpose args.
#ifdef SUNPERF
#  include <sunperf.h>
#  undef  HAVE_CBLAS
#  undef  HAVE_CLAPACK
#elif defined(INTEL_MKL)
#  include <mkl_cblas.h>
#  define HAVE_CBLAS
#  include <mkl_lapack.h>
#  undef  HAVE_CLAPACK
#elif defined(ACML)
#  include <acml.h>
#  undef  HAVE_CBLAS
#  undef  HAVE_CLAPACK
#else // use ATLAS by default
#  include <cblas.h>
#  define HAVE_CBLAS
#  ifdef USE_LAPACK
#    include <clapack.h>
#    define HAVE_CLAPACK
#  endif
#endif // vendor math libraries
#endif // USE_BLAS || USE_LAPACK


#define JTT 0
#define WAG 1
#define LG  2


#define TRUE 1
#define FALSE 0
#define MAXNSP 100
#define MAXDATASET 100
#define MAXLNAME 20
#define MAXLLINE 10000
#define lmot (8*sizeof(int))

#define seed 1234
#define deltax 1.e-4

#define nb_param_type 6    // nb of params for optimization
enum optimParam{ROOT, LENGTH, COVA, COVB, COVC, GaltierPI};
extern char aachar[];	// the allowed characters in sequences
#define lmin 1.e-6
#define lmax 9999.
#define frmin 1.e-6
#define frmax 0.999999
#define gammamin 0.005
#define gammamax 100.
#define covmin 1.e-6
#define covmax 100.
#define pimin 1.e-6
#define pimax 0.999999

#define GAMMA_RANGE_ITER 5
#define GAMMA_RANGE_SIZE 9
#define ERROR_THRESHOLD -1.e-9

typedef double proba[20][20];
typedef double dproba[4][20][20]; /* param type, nt1, nt2 */
typedef double like[20];

		/********* STRUCTURES *********/
typedef struct noeud{
	struct noeud *v1, *v2, *v3;	/* neighbors, v3 = parent */
	char* seq;			/* data if tip */
//	struct noeud *remain;		/* unchanged neighbor */
	double l1, l2, l3;		/* branch lengths */
//	double lremain;			/* unchanged length */
//	int alremain;			/* unchanged status */
//	double b1, b2, b3;		/* bootstrap values */
	int alive1, alive2, alive3;	/* initially 1, but 0 for null pointers (and -1 in the root).  unknown if these ever not redundant with ->v1 == NULL */
	int numerobr;			

	int nbpattern;
	int ps1size, ps2size;		/* max+1 of patternson1/2 */
	int *patternson1, *patternson2;
	int *underlyingbr;
	int depth;			/** was used by labelled_upgma() **/
	like **x;			/* likelihood (rate class, pattern) */
	like **dx;			/* lkh 1st derivative (rate class, pattern) */
	like **d2x;			/* lkh 2nd derivative (rate class, pattern) */
/* dx and d2x are for the parameter currently derived */
//	proba** p1;			/* subst. proba matrix toward v1 */
//	proba** p2;			/* subst. proba matrix toward v2 */
	double *p1, *p2;		// 20*nbcl square matrix, allocated in a single block
#define Pind(p,nbcl, cl,j, cl2,k) p[((cl*20 + j)*nbcl + cl2)*20 + k]
  dproba** dp1;			/* p1 first derivative */
  dproba** dp2;			/* p2 first derivative */
  dproba** d2p1;			/* p1 second derivative */
  dproba** d2p2;			/* p2 second derivative */
/* these probas have two additional dimensions with
respct to nhml1: initial rate class and final rate class */
//	double pdif1, pdif2;		/* proba switch rate class i to rate class j */
//	double pide1, pide2;		/* proba no switch */
        double** rr1, **rr2;		/* rij */
	double** drr1, **drr2;		/* rij derivatives with respect */
        double** d2rr1, **d2rr2;		/* to x=l.covar */
//	double ***x2sites;		/* lkh of the current two sites (rate class, state1, state2) */
	char nom[MAXLNAME+1];		/* name */
} *noeud;


typedef struct branche{
	noeud bout1;
	noeud bout2;
	double length;
	double bootstrap;
} *branche;


typedef struct tree{
	noeud* node;			/* array of nodes */
	int nbseq;			/* number of compared sequences */
	int lgseq;			/* length of compared sequences (patterns) */
	char** seq;			/* sequences */
	char** names;			/* sequence names */
        int* sitetopatt; 		/* site to "pattern" correspondence */
	double** dist;			/* distances between sequences */
	double* weight;			/* weights at each site */
	double weightsum;		/* sum of weights */
	double froot;			/* root->l1 / root->l2 ratio */
	double gamma_shp;
        double cov1, cov2, cov3;        // corresponding to s01, s10, s11
        double pi;                     // Galtier's pi for proportion of covarions sites 
	branche* listbr;		/* list of branches */
	int* alivebr;			/* status of branches */
	int nbalivebr;
	double* lgbr_init;		/* list of initial branch lengths */
	int nb_var_param;		/* number of optimized parameters */
	int which_var_param[6]; /* 0:fr, 1:length, 2:cov1, 3:cov2, 4.cov3,5.pi */
	int* param_nature;		/* 0:fr, 1:length, 2:cov1, 3:cov2, 4.cov3,5.pi */		
	double** var_param;		/* array of addresses of optimized parameters */
	double** classlike_s;		/* class-dependant likelihood at each site */
	double* like_s;			/* likelihood at each site (site) */
        double* likeasrv_s;
//	double*** classdlike_s;		/* class-dependant derivative at each site */
	double** dlike_s;		/* derivatives of likelihood at each site (param, site) */
	double** dlikeasrv_s;
//	double*** classd2like_s;	/* class-dependant second derivative at each site */
//	double** d2like_s;		/* second deriv. of like. at each site (param, site) */
//	double** d2likeasrv_s;
	double lkh;			/* total log likelihood ln(L) */
        double lssrv, lasrv;
	double* deriv;			/* ln(L) derivatives (param) */
	double* deriv2;			/* ln(L) second and cross derivatives (param) */
} tree;


typedef struct options{

  struct compute_option{
        int OPTIMIZE_TOPOLOGY;
	int OPTIMIZE_LENGTH;
	int OPTIMIZE_COMP;
	int OPTIMIZE_TITV;
    	int OPTIMIZE_ROOT;
	int OPTIMIZE_ANC;
	int OPTIMIZE_GAMMA;
	int OPTIMIZE_COV1;
	int OPTIMIZE_COV2;
	int OPTIMIZE_COV3;
	int OPTIMIZE_PI;
  } compute;

  struct init_option{
	double INIT_TITV;
	double INIT_ROOT;
	double INIT_ANC;
	double INIT_GAMMA;
	double INIT_COV;
	double INIT_COV1;
	double INIT_COV2;
	double INIT_COV3;
        double INIT_PI;
	int GAMMA_NCL;
	int NBRANDOM;
	int NOISE;
	char INIT_LENGTH[10];
  } init;

  struct converge_option{
	int PRECISION;
	int MAXITERATION;
	int SIMPLEX;
  } converge;

  struct print_option{
	int COV_SITE_PATTERN;	/* details about site-specific covariation */
	int OUTPUT_COMP;
	int EVAL_OUT;		/* detailed results in file detailed_out */
  } print;

  /* for program SHAKE */
  int SH_G;		
  double SH_MAXLCROSSED;
  int SH_MAXBOOTCROSSED;
} procov_options;



/***** Functions ******/
static inline void *check_alloc(size_t nbrelt, size_t sizelt){
  void *retval = calloc(nbrelt,sizelt);
  if (retval) return retval;
  perror("Not enough memory"); exit(1);
}


// procov.c
void set_var_param(tree* s_tree);
void alloc_node(noeud nd, int lgseq, int nbparam, int nbseq, int leaf, int cross, int nbcl);
void free_tree(tree* s_tree, int nbcl);
void free_node(noeud nd, int nbcl);



// mycompute_ml.c
static double minus_log_likelihood(double *paramprov);
int matmat(int nbl1, int nbc1, double *m1, int nbl2, int nbc2, double *m2, double *prod);
int invmat(int n, double *mat, double *invmat);
void reset_bl(tree* s_tree, noeud nd);
void organize_listbr(tree* s_tree);
void setpattern(tree* s_tree);
void set_numerobr(noeud nd, branche* listbr, int nbbr);
double compute(tree* s_tree, char *status);
noeud create_node(noeud v1, noeud v2, noeud v3, double l1, double l2, double l3, double b1, double b2, double b3, char* nom);


//likenode.c
void Covarionlike_node(noeud root);
void ASRVlike_node(noeud nd);

// simplex.c
double *vector(int nl, int nh);
void free_vector(double *v, int nl, int nh);
void amoeba_cont(double **p, double y[], int ndim, double ftol,
	double (*funk)(double []), int *nfunk, double** contraintes);

// readInfile.c
int readSequencefile(char *, char ***seqp, char ***seqnamep, int interleaved);


// treeio.c
noeud* ctos(char* ctree, int* notu, int *tree_is_rooted); // char* to struct
int ctot(char *input, int *arbre[], double *lgbi, double *lgbp, double* bootstrap, char *nom[], char *racine, int* nbbi);
void stoc(noeud* arbre_s, int racine, int notu, char* ctree); // struct to char*
int ctob(char *carbre, char *nom[], int *barbre[]);
void output_tree_comp(noeud nd, FILE* out, tree* s_tree); // debug dump
void pc_treeprint(const noeud p);
void makelistbr_unrooted(noeud from, noeud nd, branche* br, int* alivebr, int nbbranch);
void organize_tree(noeud from, noeud nd);
void init_bl(tree* stree);
int testbit(int *plist, int num);
int isbranch(branche br, int l1, char** list1, int l2, char** list2);
int setunderlyingbr(noeud nd, branche br, int i); 
int unroot(int** treer, int notu, double* lgbi, double* lgbp, double* bootvals, char** nom, char** list1, char** list2, int* l1, int* l2, double* fracroot1);

//shake.c

int samename(char* name1, char* name2);

// eigen.c
int eigen(int job, double A[], int n, double rr[], double ri[], double vr[], double vi[], double w[]);

// gamma.c
int DiscreteGamma(double rK[], double alpha, double beta, int K, int median);
void init_gamma(double gamma_shp, double *gamma_clmean);


// Tree searching
noeud findroot(noeud nd, int taxa);  // find root, or NULL if unrooted (no root found after 2*taxa iterations).

// aamatrix.c
void precomputeProtMat (int model);


#define maxi max
#define mini min
#define pos(i,j,n)      ((i)*(n)+(j))

/* Global variables */
int do_treesearch; // enable tree searching (set at the command line)

int do_tree_likeonly; // just print the likelihood on stdout.  for child procov
int do_tree_maxjobs;

int debug, verbose, loglevel;
FILE *out;  // the log file.
procov_options options;

tree* s_tree;
noeud root;
int gamma_nbcl, gamma_nbcl1;   // gamma_nbcl1 = gamma_nbcl*2
double freqaa[20], eigmat[20], eigenvect[20][20], probmat[20][20];
double pi, covar1, covar2, covar3; 
